/*
 * Copyright 2010 Semafor Informatik & Energie AG, Basel, Switzerland
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.

 */
package ch.semafor.gendas.dao;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;

import org.hibernate.Criteria;
import org.hibernate.Session;
import org.hibernate.stat.Statistics;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.orm.ObjectRetrievalFailureException;

/**
 * This class serves as the Base class for all other DAOs - namely to hold
 * common CRUD methods that they might all use. You should only need to extend
 * this class when your require custom CRUD logic.
 * 
 * @author <a href="mailto:tar@semafor.ch">Ronald Tannerl</a>
 * @param <T>
 *          a type variable
 * @param <PK>
 *          the primary key for that type
 */
public class GenericDaoJpa<T, PK extends Serializable> implements
    GenericDao<T, PK> {
  /**
   * Log variable for all child classes.
   */
  final private Logger log = LoggerFactory.getLogger(getClass());

  private Class<T> persistentClass;

  @PersistenceContext
  private EntityManager em; // NOPMD by wim on 9/20/10 9:00 AM

  /**
   * Constructor that takes in a class to see which type of entity to persist
   * 
   * @param persistentClass
   *          the class type you'd like to persist
   */
  public GenericDaoJpa(final Class<T> persistentClass) {
    this.persistentClass = persistentClass;
  }

  /**
   * {@inheritDoc}
   */
  @SuppressWarnings("unchecked") // NOPMD by wim on 9/20/10 9:03 AM
  public List<T> getAll() {
    return em.createQuery("from " + persistentClass.getSimpleName())
        .getResultList();
  }

  /**
   * {@inheritDoc}
   */
  @SuppressWarnings("unchecked")
  public List<T> getAllDistinct() {
	Collection result = new LinkedHashSet(getAll());
    return new ArrayList(result);
  }

  /**
   * {@inheritDoc}
   */
  @SuppressWarnings("unchecked")
  public T get(final PK id) { // NOPMD by wim on 9/20/10 9:00 AM
    T entity = (T) em.find(this.persistentClass, id);

    if (entity == null) {
      log.warn("Uh oh, '" + this.persistentClass + "' object with id '" + id
          + "' not found...");
      throw new ObjectRetrievalFailureException(this.persistentClass, id);
    }

    return entity;
  }

  /**
   * {@inheritDoc}
   */
  @SuppressWarnings("unchecked")
  public boolean exists(final PK id) { // NOPMD by wim on 9/20/10 9:00 AM
	T entity = this.get(id);
    return entity != null;
  }

  /**
   * {@inheritDoc}
   */
  @SuppressWarnings("unchecked")
  public T save(final T object) {
    return (T) em.merge(object);
  }

  /**
   * {@inheritDoc}
   */
  public void remove(final PK id) { // NOPMD by wim on 9/20/10 9:00 AM
	T entity = this.get(id);
    em.remove(entity);
  }

  /**
   * {@inheritDoc}
   */
  @SuppressWarnings("unchecked")
  public List<T> findByNamedQuery(final String queryName,
      final Map<String, Object> queryParams) {
	Query q = em.createNamedQuery(queryName); // NOPMD by wim on 9/20/10 9:00 AM
    for (String key : queryParams.keySet()) {
      q.setParameter(key, queryParams.get(key));
    }
    return q.getResultList();
  }

  /**
   * {@inheritDoc}
   */
  @SuppressWarnings("unchecked")
  public int countByNamedQuery(final String queryName,
      final Map<String, Object> queryParams) {
    Query q = em.createNamedQuery(queryName); // NOPMD by wim on 9/20/10 9:03 AM
    for (String key : queryParams.keySet()) {
      q.setParameter(key, queryParams.get(key));
    }
    return ((Long)(q.getResultList().get(0))).intValue();
  }

  public EntityManager getEntityManager() {
    return this.em;
  }

  public Criteria createCriteria(Class clazz, String alias) {
	    return ((Session) em.getDelegate()).createCriteria(clazz, alias);
	  }
  public Criteria createCriteria(Class clazz) {
	    return ((Session) em.getDelegate()).createCriteria(clazz);
	  }
  public Criteria createCriteria() {
	    return createCriteria(this.persistentClass);
	  }
  
  public Statistics getStatistics(){
	  return ((Session)em.getDelegate()).getSessionFactory().getStatistics();
  }
}