/*
 * Copyright 2010 Semafor Informatik & Energie AG, Basel, Switzerland
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.

 */
package ch.semafor.gendas.model;

import java.io.Serializable;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@Entity
@Table(name = "properties")
public class Property implements Serializable {
	@Transient
	private final Logger logger = LoggerFactory.getLogger(Property.class); 
	private static final long serialVersionUID = 1L;

	@Id
	@GeneratedValue(strategy = GenerationType.AUTO)
	private Long id; // NOPMD by wim on 9/20/10 2:07 PM

	@OneToMany(mappedBy = "property", cascade = CascadeType.ALL)
	private List<PropertyValueList> valuelist = new ArrayList<PropertyValueList>(); 

	@ManyToOne
	@JoinColumn(name = "element_id")
	private Element element;

	@ManyToOne(cascade = CascadeType.MERGE)
	@JoinColumn(name = "property_type_id")
	private PropertyType type;

	@Column(nullable = false)
	private Long revision;

	@Column(nullable = false)
	private Long nextRevision;

	// Constructor
	public Property() {
		super();
		this.element = null;
	}

	// Constructor
	public Property(final Element el) throws CoreException { 
		super();
		this.id = null;
		this.element = el;
		this.revision = el.getNewRevision();
		this.nextRevision = Modification.MaxRevision;
		el.addProperty(this);
	}

	// Constructor
	public Property(Element el, PropertyType type) throws CoreException { 
		super();
		this.id = null;
		this.type = type;
		this.element = el;
		this.revision = el.getNewRevision();
		this.nextRevision = Modification.MaxRevision;
		el.addProperty(this);
	}

	public String getName() {
		if (this.type != null) {
			return this.type.getName();
		}
		return "<no type>";
	}

	public boolean isInRevision(final long rev) {
		if (rev == Modification.MaxRevision) {
			return this.nextRevision == Modification.MaxRevision; 
		}
		return rev >= this.revision && rev < this.nextRevision;
	}

	// getter and setter of Revision
	public void setRevision(final long rev) {
		this.revision = rev;
	}

	public Long getRevision() {
		return this.revision;
	}

	// getter and setter of NextRevision
	public void setNextRevision(final long rev) {
		this.nextRevision = rev;
	}

	public Long getNextRevision() {
		return this.nextRevision;
	}

	public void initRevision(final long rev) throws CoreException {
		if (isPersistent()) {
			throw new CoreException("Property has to be transient");
		}
		this.revision = rev;
		this.nextRevision = Modification.MaxRevision;
		if (this.valuelist.size() > 1) {
			throw new CoreException(
					"Only one ValueList allowed in transient Properties");
		}
		for (PropertyValueList lst : this.valuelist) {
			lst.setRevision(rev);
			lst.setNextRevision(this.nextRevision);
		}
	}

	// Get the latest available Revision number.
	public long getLastRevision() throws CoreException {
		return this.element.getLastRevision();
	}

	// New Revision for current Transaction
	public long getNewRevision() throws CoreException {
		return this.element.getNewRevision();
	}

	public boolean isLastRevision() {
		return this.nextRevision == Modification.MaxRevision;
	}

	public void addValueList(final PropertyValueList vlist)
			throws CoreException {
		if (this != vlist.getProperty()) {
			throw new CoreException("Property #" + this.getId()
					+ " is not attached to this Element, but #"
					+ vlist.getProperty().getId() + " is");
		}
		for (PropertyValueList lst : this.valuelist) {
			if (lst == vlist) {
				throw new CoreException("ValueList already added to Property");
			}
		}
		this.valuelist.add(vlist);
	}

	public PropertyValueList getValueList(final Long revision) {
		int inx = this.valuelist.size() - 1;
		PropertyValueList vlist;
		while (inx >= 0) {
			vlist = this.valuelist.get(inx);
			if (vlist.isInRevision(revision)) {
				return vlist; 
			}
			inx--;
		}
		return null;
	}

	private PropertyValueList getLastValueList() {
		for (PropertyValueList lst : this.valuelist) {
			if (lst.isLastRevision()) {
				return lst; 
			}
		}
		return null;
	}

	public String getString(final int index) {
		final PropertyValueList vlist = getValueList(Modification.MaxRevision);
		if (vlist == null) {
			return null; 
		}
		return vlist.getString(index);
	}

	public Double getDouble(final int index) {
		final PropertyValueList vlist = getValueList(Modification.MaxRevision);
		if (vlist == null) {
			return null; 
		}
		return vlist.getDouble(index);
	}

	public Integer getInt(final int index) {
		final PropertyValueList vlist = getValueList(Modification.MaxRevision);
		if (vlist == null) {
			return null; 
		}
		return vlist.getInt(index);
	}

	public Long getLong(final int index) {
		final PropertyValueList vlist = getValueList(Modification.MaxRevision);
		if (vlist == null) {
			return null; 
		}
		return vlist.getLong(index);
	}

	public BigDecimal getDecimal(final int index) {
		final PropertyValueList vlist = getValueList(Modification.MaxRevision);
		if (vlist == null) {
			return null; 
		}
		return vlist.getDecimal(index);
	}

	public GregorianCalendar getDate(final int index) {
		final PropertyValueList vlist = getValueList(Modification.MaxRevision);
		if (vlist == null) {
			return null; 
		}
		return vlist.getDate(index);
	}

	public void setValue(final Long revision, final int index,
			final PropertyValue value) throws CoreException {
		PropertyValueList vlist = getValueList(revision);
		if (vlist == null) {
			vlist = new PropertyValueList(this);
		}
		vlist.setValue(index, value);
	}

	public void setString(final int index, final String val)
			throws CoreException {
		final PropertyValue value = new PropertyValue(null, val);
		setValue(getLastRevision(), index, value);
	}

	public void setDouble(final int index, final Double val)
			throws CoreException {
		final PropertyValue value = new PropertyValue(null, val);
		setValue(getLastRevision(), index, value);
	}

	public void setInt(final int index, final Integer val) throws CoreException {
		final PropertyValue value = new PropertyValue(null, val);
		setValue(getLastRevision(), index, value);
	}

	public void setDecimal(final int index, final BigDecimal val) throws CoreException {
		final PropertyValue value = new PropertyValue(null, val);
		setValue(getLastRevision(), index, value);
	}

	public void setLong(final int index, final Long val) throws CoreException {
		final PropertyValue value = new PropertyValue(null, val);
		setValue(getLastRevision(), index, value);
	}

	public void setDate(final int index, final GregorianCalendar val)
			throws CoreException {
		final PropertyValue value = new PropertyValue(null, val);
		setValue(getLastRevision(), index, value);
	}

	public void setBool(final int index, final Boolean val)
			throws CoreException {
		final PropertyValue value = new PropertyValue(null, val);
		setValue(getLastRevision(), index, value);
	}

	// getter and setter of Element
	public Element getElement() {
		return element;
	}

	public void setElement(Element el) {
		this.element = el;
	}

	// getter and setter of Type
	public PropertyType getType() {
		return type;
	}

	public void setType(final PropertyType type) {
		this.type = type;
	}

	// getter and setter of Id
	public Long getId() {
		return id;
	}

	public void setId(Long id) { 
		this.id = id;
	}

	public boolean isId(long id) { 
		if (this.id != null) {
			if (this.id.longValue() == id) { 
				return true; 
			}
		}
		return false;
	}

	public boolean isTransient() {
		return this.id == null;
	}

	public boolean isPersistent() {
		return this.id != null;
	}

	public String toString() {
		ToStringBuilder sb = new ToStringBuilder(this,
				ToStringStyle.SHORT_PREFIX_STYLE); 
		sb.append("id", this.id);
		sb.append("type", this.type.getName());
		sb.append("revision", revision + "-" + nextRevision);
		for (PropertyValueList v : valuelist) {
			sb.append(v.toString());
		}
		return sb.toString();
	}

	public void print(final int ind) {
		ToStringBuilder sb = new ToStringBuilder(this,
				ToStringStyle.SHORT_PREFIX_STYLE); 
		sb.append("id", this.id);
		sb.append("serialVersionUID", serialVersionUID);
		sb.append("revision", revision + "-" + nextRevision);
		sb.append("type", this.type.toString());
		element.indent(ind);
		System.out.println(sb.toString()); 
		for (PropertyValueList vlist : this.valuelist) {
			vlist.print(ind + 1);
		}
	}

	public boolean isValid() {
		final PropertyValueList vlist = getLastValueList();
		if (vlist == null) {
			return false; 
		}
		return vlist.isValid();
	}

	public boolean isntValid() {
		return !isValid();
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean equals(Object o) { 
		if (this == o) {
			return true; 
		}
		if (!(o instanceof Property)) {
			return false; 
		}
		final Property p = (Property) o; 
		if (type != null) { 
			if (!type.equals(p.type)) {
				return false; 
			}
		} else if (p.type == null) {
			return true; 
		}
		// logger.debug("compare values " + values.size() + "==" +
		// p.values.size());
		final PropertyValueList this_vlist = getValueList(Modification.MaxRevision);
		final PropertyValueList that_vlist = p
				.getValueList(Modification.MaxRevision);

		if (this_vlist == null) {
			if (that_vlist != null && !that_vlist.isEmpty()) {
				return false; 
			}
			// logger.debug("no value list");
			return true; 
		}
		if (that_vlist == null) {
			return false; 
		}
		return this_vlist.equals(that_vlist);
	}

	/**
	 * {@inheritDoc}
	 */
	public int hashCode() {
		final int prime = 31; 
		int result = 1;
		result = prime * result + (type != null ? type.hashCode() : 0); 
		return result;
	}

	public void assign(Property from) throws CoreException { 
		logger.debug("begin of Property::assign(): to {}", this.getName());
		if (type == null || from.type == null) {
			throw new CoreException(
					"Properties without a type not possible in assign()");
		}
		if (type != from.type) {
			throw new CoreException(
					"A change of PropertyType is not allowed in assign()");
		}
		final PropertyValueList src_vlist = from.getLastValueList();
		if (src_vlist == null) {
			logger.debug("no values available => delete");
			delete();
		} else if (src_vlist.isntValid()) {
			logger.debug("no valid values available in {}: => delete",
					from.getName());
			delete();
		} else {
			logger.debug("Source Property : {}", src_vlist.toString());
			PropertyValueList dest_vlist = getLastValueList();
			if (dest_vlist == null) {
				logger.debug("create new PropertyValueList");
				dest_vlist = new PropertyValueList(this);
			}
			dest_vlist.assign(src_vlist);
			this.nextRevision = Modification.MaxRevision;
		}
		logger.debug("end of Property::assign()");
	}

	public void delete() throws CoreException {
		logger.debug("delete Property {}", this.toString());

		if (this.nextRevision != Modification.MaxRevision) {
			logger.debug("already deleted");
		} else {
			// set invalid
			this.nextRevision = getNewRevision();
			final PropertyValueList vlist = getLastValueList();
			if (vlist.isLastRevision()) {
				vlist.setNextRevision(this.nextRevision);
			}
		}
	}

	public boolean checkParent(Element e) {
		return this.element != e;
	}

	public List<PropertyValueList> getValuelist() {
		return this.valuelist;
	}

}
