/*
 * Copyright 2010 Semafor Informatik & Energie AG, Basel, Switzerland
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.

 */
package ch.semafor.gendas.service;

import java.util.List;
import java.util.Map;

import net.sf.ehcache.Statistics;

import ch.semafor.gendas.model.CoreException;
import ch.semafor.gendas.model.Element;
import ch.semafor.gendas.model.ElementType;
import ch.semafor.gendas.model.Modification;
/**
 * Service Interface to handle all persistence operations
 * @author tar
 *
 */
public interface PersistenceManager {
    /**
     * creates a new Element Type with corresponding Property Types
     * 
     * @param beanClass
	 * @param idName
	 *            name of key value
	 * @param idVersion
	 *            name of version field for optimistic locking
     * @return created element type
     */
	@SuppressWarnings("rawtypes")
	// NOPMD by wildi on 9/21/10 6:26 AM
	ElementType createElementType(Class beanClass, String idName, String idVersion);

	/**
	 * get all elements
	 * 
	 * @return list of elements
	 */
	List<Element> getAllElements();

	/**
	 * get all element types
	 * 
	 * @return list of element types
	 */
	List<ElementType> getAllElementsTypes();

	/**
	 * insert or update bean
	 * 
	 * @param bean
	 *            to be persisted
	 * @return persisted element
	 * @throws CoreException
	 * @throws ElementCreationException
	 */
	Element save(final Object bean )
			throws CoreException, ElementCreationException;

	/**
	 * load bean by id
	 * 
	 * @param id
	 *            key value of element
	 * @return loaded bean
	 */
	Object load(Long id)
			throws CoreException, ElementCreationException; // NOPMD by wildi on
															// 9/21/10 6:26 AM

	/**
	 * load bean by id and revision
	 * 
	 * @param id
	 *            key value of element
	 * @param revision
	 * @return loaded bean
	 */
	Object load(Long id, Long revision) throws CoreException,
			ElementCreationException; // NOPMD by wildi on 9/21/10 6:26 AM

	/**
	 * get list of modifications
	 * @param id of element
	 * @return list of modifications
	 */
	List<Modification> getModifications(Long id); // NOPMD by wildi on 9/21/10 6:26 AM

	/**
	 * get revision of last modification
	 * @param id of element
	 * @return revision
	 */
	Long getLastRevision(Long id); // NOPMD by wildi on 9/21/10 6:26 AM

	/**
	 * create list of persisted beans of beanClass with property value and set
	 * their id property
	 * 
	 * @param beanClass
	 * @param args
	 * 			  map of property name and value pairs
	 * @return list of persisted objects
	 * 
	 * @throws ElementCreationException
	 */
	// ToDo, PMD:LooseCoupling
	@SuppressWarnings("rawtypes")
	List findDomainObjectsByArgs(final Class beanClass, final Map<String, String> args,
			final Map<String, Map<String, String>> childargs)
			throws ElementCreationException; // NOPMD by wildi on 9/21/10 6:26
												// AM

	/**
	 * get elements by class, prop and value
	 * 
	 * @param beanClass
	 * @param prop
	 * @param val
	 * @return list of elements
	 */
	List getAllElementsByPropertyValue(final Class beanClass,
			final String prop, final String val);

	/**
	 * get objects by class, prop and value
	 * 
	 * @param beanClass
	 * @param prop
	 * @param val
	 * @return list of beans
	 */
	@SuppressWarnings("rawtypes")
	List getAllObjectsByPropertyValue(final Class beanClass, final String prop,
			final String val)
			throws ElementCreationException; // NOPMD by wildi on 9/21/10 6:26
												// AM

	/**
	 * get count of beanClass by or-ing 2 property values
	 * 
	 * @param beanClass
	 * @param prop1
	 *            , prop2 name of property
	 * @param val1
	 *            , val2 value of property
	 * @return number of objects
	 */
	@SuppressWarnings("rawtypes")
	int getCountOfObjectsBy2PropertyValues(final Class beanClass,
			final String prop1, final String val1, final String prop2,
			final String val2);

	/**
	 * get count of beanClass with property value
	 * 
	 * @param beanClass
	 * @param prop
	 *            name of property
	 * @param val
	 *            value of property
	 * @return number of objects
	 */
	@SuppressWarnings("rawtypes")
	int getCountOfObjectsByPropertyValue(final Class beanClass,
			final String prop, final String val);

	/**
	 * return number of persisted objects of beanClass with given properties
	 * 
	 * @param beanClass
	 * @param args
	 *            map of property name and value pairs of this root element
	 * @param childargs
	 * @return list of persisted objects
	 * @throws ElementCreationException
	 */
	// ToDo, PMD:LooseCoupling
	@SuppressWarnings("rawtypes")
	int getCountDomainObjectsByArgs(final Class beanClass, final Map<String, String> args,
			final Map<String, Map<String,String>> childargs);

	/**
	 * create list of persisted objects of beanClass with reference bean ref and
	 * set their id property
	 * 
	 * @param beanClass
	 * @param ref
	 *            reference bean
	 * @return list of persisted objects
	 * 
	 * @throws ElementCreationException
	 */
	@SuppressWarnings("rawtypes")
	List getAllObjectsByReference(final Class beanClass, final Object ref) throws ElementCreationException;

	/**
	 * create list of persisted objects of beanClass and set their id property
	 * 
	 * @param beanClass
	 * @return list of persisted objects
	 * 
	 * @throws ElementCreationException
	 */
	@SuppressWarnings("rawtypes")
	List getAllObjects(final Class beanClass) throws ElementCreationException; 
	
	/**
	 * delete object (do not actually deleted it but mark it as deleted)
	 * @param id of object to be deleted
	 */
	void delete(Long id) throws CoreException; 

	/**
	 * get element by id
	 * @param id of element
	 * @return found element if any 
	 */
	Element getElement( Long id );
	
	/**
	 * switch caching on or off for load operation
	 * @param flag
	 */
	void activateCaching( boolean flag );
	
	/**
	 * get cache statistics
	 * @return cache statistics
	 */
	Statistics getCacheStatistics();
	
}
